from magma import *
from mantle import Not
from .register import _RegisterName, FFs

__all__ = ['DefineJohnson', 'Johnson']

def DefineJohnson(n, has_ce=False, has_reset=False):
    """
    Generate a n-bit johnson counter.

    Walking ring counter / Moebios counter

    O : Out(Bits(n))
    """
    class Johnson(Circuit):
        name = _RegisterName('Johnson', n, 0, has_ce, has_reset)
        IO = ['O', Out(Bits[ n ])] + ClockInterface(has_ce,has_reset)
        @classmethod
        def definition(johnson):
            ffs = FFs(n, has_ce=has_ce, has_reset=has_reset)
            reg = scan(ffs, scanargs={"I":"O"})
            johnson.O(reg(Not()(reg.O[n-1])))
            wireclock(johnson, reg)
    return Johnson

def Johnson(n, has_ce=False, has_reset=False, **kwargs):
    return DefineJohnson(n, has_ce, has_reset)(**kwargs)



import magma as m
from .LUT import LUT
from .compare import EQ

__all__  = ['DefineDecode', 'Decode', 'decode']

@m.cache_definition
def DefineDecode(i, n):
    circ = m.DefineCircuit(f"Decode{n}_{i:X}", "I", m.In(m.Bits[ n ]), "O", m.Out(m.Bit))
    m.wire(circ.O, EQ(n)(circ.I, m.bits(i, n)))
    m.EndDefine()
    return circ

def Decode(i, n, invert=False, **kwargs):
    """
    Decode the n-bit number i.

    @return: 1 if the n-bit input equals i
    """

    if n <= 4:
        i = 1 << i
        if invert:
            mask = (1 << (1 << n)) - 1
            i = mask & (~i)
        return m.uncurry(LUT(i, n, **kwargs))

    return DefineDecode(i, n)()

def decode(I, i, invert=False, **kwargs):
    return Decode(i, len(I), invert=invert, **kwargs)(I)




from magma import *

SB_IO= DeclareCircuit("SB_IO", 
                        'PACKAGE_PIN', In(Bit), # should be inout
                        'CLOCK_ENABLE', In(Bit),
                        'INPUT_CLOCK', In(Bit),
                        'OUTPUT_CLOCK', In(Bit),
                        'OUTPUT_ENABLE', In(Bit),
                        'LATCH_INPUT_VALUE', In(Bit),
                        'D_IN_0', In(Bit), # rising
                        'D_IN_1', In(Bit), # falling
                        'D_OUT_0', Out(Bit), # rising
                        'D_OUT_1', Out(Bit) # falling
                     )

#module top (input PIO1_02, output D1);
#wire y;
#SB_IO #(.PIN_TYPE(6'b00_0000)) in  ( .PACKAGE_PIN(PIO1_02), .D_IN_0(y) );
#SB_IO #(.PIN_TYPE(6'b01_0110)) out ( .PACKAGE_PIN(D1), .D_OUT_0(y) );
#endmodule




from magma import *
from mantle import And, DefineAdd
from mantle import Mux
from mantle import Add
from mantle import Register
from mantle import Decode
import math

__all__  = ['DefineCounter', 'Counter']
__all__ += ['DefineUpCounter', 'UpCounter']
__all__ += ['DefineDownCounter', 'DownCounter']
__all__ += ['DefineUpDownCounter', 'UpDownCounter']


def _CounterName(name, incr, has_ce, has_reset, cin, cout):
    if incr != 1: name += f'_{incr}'
    if has_ce: name += 'CE'
    if has_reset:  name += 'R'
    if cin: name += '_CIN'
    if cout: name += '_COUT'
    return name

#
# Create an n-bit counter with a given increment.
#
# O : Out(UInt(n)), COUT : Out(Bit)
#
def DefineCounter(n, cin=False, cout=True, incr=1, has_ce=False, has_reset=False):

    name = _CounterName(f'Counter{n}', incr, has_ce, has_reset, cin, cout)

    args = []
    if cin:
        args += ['CIN', In(Bit)]

    args += ["O", Out(UInt[ n ])]
    if cout:
        args += ["COUT", Out(Bit)]

    args += ClockInterface(has_ce, has_reset)

    Counter = DefineCircuit(name, *args)

    add = DefineAdd(n, cin=cin, cout=cout)()
    reg = Register(n, has_ce=has_ce, has_reset=has_reset)

    wire( reg.O, add.I0 )
    wire( array(incr, n), add.I1 )

    reg(add.O)

    next = False
    if next:
        wire( add.O, Counter.O )
    else:
        wire( reg.O, Counter.O )

    if cin:
        wire( Counter.CIN, add.CIN )

    if cout:
        wire( add.COUT, Counter.COUT )

    wireclock(Counter, reg)
    wiredefaultclock(Counter, reg)

    EndCircuit()

    return Counter

def Counter(n, cin=False, cout=True, incr=1,
             has_ce=False, has_reset=False, **kwargs):
    """Construct a n-bit up counter."""
    return DefineUpCounter(n, cin=cin, cout=cout, incr=incr,
               has_ce=has_ce, has_reset=has_reset)(**kwargs)

DefineUpCounter = DefineCounter
UpCounter = Counter

#
# Create an n-bit down counter.
#
# O : Out(UInt(n)), COUT : Out(Bit)
#
def DefineDownCounter(n, cin=False, cout=True, decr=1,
    has_ce=False, has_reset=False):
    incr = (1 << n) - (decr)
    return DefineCounter(n, cin=cin, cout=cout, incr=incr,
               has_ce=has_ce, has_reset=has_reset)

def DownCounter(n, cin=False, cout=True, decr=1,
    has_ce=False, has_reset=False, **kwargs):
    return DefineDownCounter(n, cin=cin, cout=cout, decr=decr,
               has_ce=has_ce, has_reset=has_reset)(**kwargs)


#
# Create an n-bit up-down counter.
#
# U : In(Bit), D : In(Bit), O : Out(UInt(n)), COUT : Out(Bit)
#
def DefineUpDownCounter(n, cout=True, has_ce=False, has_reset=False):

    name = _CounterName(f'UpDownCounter{n}', 1, has_ce, has_reset, False, cout)

    args = []

    args += ["U", In(Bit)]
    args += ["D", In(Bit)]

    args += ["O", Out(UInt[ n ])]
    if cout:
        args += ["COUT", Out(Bit)]

    args += ClockInterface(has_ce, has_reset)

    Counter = DefineCircuit(name, *args)

    add = Add(n, cin=True, cout=cout)
    reg = Register(n, has_ce=has_ce, has_reset=has_reset)

    wire( reg.O, add.I0 )
    wire( array(n*[Counter.D]), add.I1 )
    wire( Counter.U, add.CIN )

    reg(add)

    next = False
    if next:
        wire( add.O, Counter.O )
    else:
        wire( reg.O, Counter.O )

    if cout:
        wire( add.COUT, Counter.COUT )

    wireclock(Counter, reg)

    EndCircuit()

    return Counter

def UpDownCounter(n, cout=True, has_ce=False, has_reset=False, **kwargs):
    """Construct an n-bit updown counter."""
    return DefineUpDownCounter(n, cout=cout,
              has_ce=has_ce, has_reset=has_reset)(**kwargs)


def DefineCeilFloorUpDownCounter(m, has_ce=False, has_reset=False):
    """
    Counter between 0 and m - 1 that uses the minimum number of bits.
    
    This counter counts up by 1 if U is True and D is False, down by 1 if U is False and D is True,
    no change if both are False or True
    
    Also, unlike normal updown counter, up by 1 has no effect if already at m - 1 and
    down by 1 has nefect if already at 0.
    
    :param m: The value the counter counts up to
    :param has_ce: Whether this counter should a clock-enable input
    :return: An updown counter circuit
    """
    class CeilFloorUpDownCounter(Circuit):
        num_bits = math.ceil(math.log(m, 2))
        name = "CeilFloorUpDownCounter_m{}_hasCE{}_hasReset{}".format(str(m), str(has_ce), str(has_reset))
        IO = ['U', In(Bit), 'D', In(Bit), 'O', Out(UInt[num_bits])] + ClockInterface(has_ce, has_reset)

        @classmethod
        def definition(ceilFloorUpDownCounter):
            up_down_counter = UpDownCounter(ceilFloorUpDownCounter.num_bits, False, has_ce, has_reset)
            at_max = Decode(m - 1, up_down_counter.O.N)(up_down_counter.O)
            at_min = Decode(0, up_down_counter.O.N)(up_down_counter.O)

            # only increase counter if input says to increase and not at max
            should_increase = ~at_max & ceilFloorUpDownCounter.U & ~ceilFloorUpDownCounter.D
            wire(should_increase, up_down_counter.U)
            # only decrease counter if input says to decrease and not at min
            should_decrease = ~at_min & ~ceilFloorUpDownCounter.U & ceilFloorUpDownCounter.D
            wire(should_decrease, up_down_counter.D)

            wire(up_down_counter.O, ceilFloorUpDownCounter.O)

    return CeilFloorUpDownCounter

def CeilFloorUpDownCounter(m, has_ce=False, has_reset=False):
    return DefineCeilFloorUpDownCounter(m, has_ce, has_reset)()



from magma import *
from .gates import Buf as gatesbuf

__all__ = ['DefineBuf', 'Buf']

def DefineOp(name, height, op):

    if height != 1:
        name += str(height)
    args = [name]
    for i in range(height):
        args += ['I{}'.format(i), In(Bit)]
    args += ['O', Out(Bit)]

    c = DefineCircuit(*args)
    op = op(height)
    EndCircuit()
    if   height == 1:
        wire(op(c.I0), c.O)
    elif height == 2:
        wire(op(c.I0, c.I1), c.O)
    elif height == 3:
        wire(op(c.I0, c.I1, c.I2), c.O)
    elif height == 4:
        wire(op(c.I0, c.I1, c.I2, c.I3), c.O)
    return c

def DefineOpW(name, height, width, op):
    if height == 1:
        name += str(width)
    else:
        name += str(height) + 'x' + str(width)
    T = Bits(width)
    args = [name]
    for i in range(height):
        args += ['I{}'.format(i), In(T)]
    args += ['O', Out(T)]

    c = DefineCircuit(*args)
    def f(y):
        return op(height)
    op = join(col(f, width))
    EndCircuit()
    if   height == 1:
        wire(op(c.I0), c.O)
    elif height == 2:
        wire(op(c.I0, c.I1), c.O)
    elif height == 3:
        wire(op(c.I0, c.I1, c.I2), c.O)
    elif height == 4:
        wire(op(c.I0, c.I1, c.I2, c.I3), c.O)
    return c

def DefineBuf(width=None):
    if width is None:
        return DefineOp('Buf', 1, gatesbuf)
    return DefineOpW('Buf', 1, width, gatesbuf)

def Buf(width=None):
    return DefineBuf(width)()



from magma import *
from ..ice40.IOB import SB_IO

__all__ = ['IBuf', 'OBuf']

#
# SB_IO #(.PIN_TYPE(6'b00_0000)) in  ( .PACKAGE_PIN(PIO1_02), .D_IN_0(y) );
#
def IBuf(pin, **kwargs):
    io = SB_IO(**kwargs)
    wire(pin, io.PACKAGE_PIN)
    return AnonymousCircuit("O", io.D_IN_0);
    
#
# SB_IO #(.PIN_TYPE(6'b01_0110)) out ( .PACKAGE_PIN(D1), .D_OUT_0(y) );
#
def OBuf(pin, **kwargs):
    io = SB_IO(**kwargs)
    wire(pin, io.PACKAGE_PIN)
    return AnonymousCircuit("I", io.D_OUT_0);
    




from magma import *
from hwtypes import BitVector
from .arith import declare_binop, get_length
from .logic import not_, XOr
import operator
from .util import DefineCoreirCircuit, DeclareCoreirCircuit


def DefineEQ(width):
    def simulate(self, value_store, state_store):
        I0 = BitVector[width](value_store.get_value(self.I0))
        I1 = BitVector[width](value_store.get_value(self.I1))
        #O = operator.eq(I0, I1).as_bool_list()[0]
        O = int(operator.eq(I0, I1))
        value_store.set_value(self.O, O)
    if width is None:
        circ = DefineCircuit('corebit_eq', 'I0', In(Bit), 'I1', In(Bit), 'O', Out(Bit))
        wire(circ.O, not_(XOr(2, None)(circ.I0, circ.I1)))
        EndDefine()
        return circ
    else:
        T = Bits[width]
        return DeclareCoreirCircuit("coreir_eq_{}".format(width),
                              'I0', In(T), 'I1', In(T),
                              'O', Out(Bit),
                              stateful=False,
                              simulate=simulate,
                              verilog_name="coreir_eq",
                              coreir_name="eq",
                              coreir_lib = "coreir",
                              coreir_genargs={"width": width})

DefineCoreirEq = DefineEQ

def EQ(n, **kwargs):
    return DefineEQ(n)(**kwargs)


DefineCoreirNeq = declare_binop("neq", operator.ne, out_type=Bit)

def DefineNE(n):
    if n is None:
        T = Bit
    else:
        T = Bits[n]
    circ = DefineCircuit("NE{}".format(n),
        "I0", In(T), "I1", In(T), "O", Out(Bit))
    eq = EQ(n)
    O = not_(eq(circ.I0, circ.I1))
    wire(O, circ.O)
    EndDefine()
    return circ

def NE(n, **kwargs):
    return DefineNE(n)(**kwargs)


DefineCoreirUlt = declare_binop("ult", operator.lt, out_type=Bit)


def DefineULT(n):
    return DefineCoreirUlt(n)

def ULT(n, **kwargs):
    return DefineULT(n)(**kwargs)


DefineCoreirUle = declare_binop("ule", operator.le, out_type=Bit)


def DefineULE(n):
    return DefineCoreirUle(n)

def ULE(n, **kwargs):
    return DefineULE(n)(**kwargs)


DefineCoreirUgt = declare_binop("ugt", operator.gt, out_type=Bit)


def DefineUGT(n):
    return DefineCoreirUgt(n)

def UGT(n, **kwargs):
    return DefineUGT(n)(**kwargs)


DefineCoreirUge = declare_binop("uge", operator.ge, out_type=Bit)


def DefineUGE(n):
    return DefineCoreirUge(n)

def UGE(n, **kwargs):
    return DefineUGE(n)(**kwargs)


DefineCoreirSlt = declare_binop("slt", operator.lt, out_type=Bit)


def DefineSLT(n):
    return DefineCoreirSlt(n)

def SLT(n, **kwargs):
    return DefineSLT(n)(**kwargs)


DefineCoreirSle = declare_binop("sle", operator.le, out_type=Bit)


def DefineSLE(n):
    return DefineCoreirSle(n)

def SLE(n, **kwargs):
    return DefineSLE(n)(**kwargs)


DefineCoreirSgt = declare_binop("sgt", operator.gt, out_type=Bit)


def DefineSGT(n):
    return DefineCoreirSgt(n)

def SGT(n, **kwargs):
    return DefineSGT(n)(**kwargs)


DefineCoreirSge = declare_binop("sge", operator.ge, out_type=Bit)


def DefineSGE(n):
    return DefineCoreirSge(n)

def SGE(n, **kwargs):
    return DefineSGE(n)(**kwargs)



from magma import *
from magma.bitutils import log2
from mantle import *
from .swap import swaps, evenoddswaps

__all__  = ['DefineBrickSorter', 'BrickSorter', 'bricksorter']

#
# Brick Sorter
#
def DefineBrickSorter(n):
    assert n % 2 == 0
    T = Bits(n)
    class _BrickSorter(Circuit):
        name = 'BrickSorter{}'.format(n)
        IO = ['I', In(T), "O", Out(T)]
        @classmethod
        def definition(io):
            I = io.I
            for i in range(n//2):
                I = swaps(I)
                I = evenoddswaps(I)
            wire(I, io.O)
    return _BrickSorter

def BrickSorter(n):
    return DefineBrickSorter(n)()

def bricksorter(I):
    return BrickSorter(len(I))(I)



from magma import *
from hwtypes import BitVector
from magma.frontend.coreir_ import CircuitInstanceFromGeneratorWrapper, GetCoreIRBackend
from .register import register
from functools import lru_cache


def gen_sim_mem(depth, width):
    addr_width = getRAMAddrWidth(depth)
    def sim_mem(self, value_store, state_store):
        cur_clk = value_store.get_value(self.clk)

        if not state_store:
            state_store['mem'] = [
                BitVector[width](0) for _ in range(depth)
            ]
            state_store['prev_clk'] = cur_clk


        prev_clk = state_store['prev_clk']
        clk_edge = cur_clk and not prev_clk
        rdata = value_store.get_value(self.rdata)

        if clk_edge:
            index = BitVector[addr_width](value_store.get_value(self.raddr)).as_int()
            rdata = state_store['mem'][index].as_bool_list()
        if clk_edge:
            if value_store.get_value(self.wen):
                index = BitVector[addr_width](value_store.get_value(self.waddr)).as_int()
                state_store['mem'][index] = BitVector[width](value_store.get_value(self.wdata))

        state_store['prev_clk'] = cur_clk
        value_store.set_value(self.rdata, rdata)
    return sim_mem


@lru_cache(maxsize=None)
def DefineCoreirMem(depth, width):
    name = "coreir_mem{}x{}".format(depth,width)
    addr_width = getRAMAddrWidth(depth)
    IO = ["raddr", In(Bits[ addr_width ]),
          "rdata", Out(Bits[ width ]),
          "waddr", In(Bits[ addr_width ]),
          "wdata", In(Bits[ width ]),
          "clk", In(Clock),
          "wen", In(Bit) ]
    return DeclareCircuit(name, *IO, verilog_name="coreir_mem",
            coreir_name="mem", coreir_lib="coreir",
            simulate=gen_sim_mem(depth, width),
            coreir_genargs={"width": width, "depth": depth})
            # coreir_configargs={"init": "0"})

def CoreirMem(depth, width):
    return CircuitInstanceFromGeneratorWrapper(GetCoreIRBackend(), "coreir", "mem", "CoreIRmem_w{}_d{}".format(width, depth),
                                               ["global"], {"width": width, "depth": depth})

def DefineROM(height, width,read_latency=0):
    """
    coreir doesn't have a ROM primitive yet
    """
    raise NotImplementedError()

def getRAMAddrWidth(height):
    return max((height - 1).bit_length(), 1)

@lru_cache(maxsize=None)
def DefineRAM(height, width, read_latency=0):
    addr_width = getRAMAddrWidth(height)
    circ = DefineCircuit("RAM{}x{}".format(height, width),
        "RADDR", In(Bits[ addr_width ]),
        "RDATA", Out(Bits[ width ]),
        "WADDR", In(Bits[ addr_width ]),
        "WDATA", In(Bits[ width ]),
        "CLK",   In(Clock),
        "WE",    In(Bit),
    )
    coreir_mem = DefineCoreirMem(height, width)()
    wire(circ.RADDR, coreir_mem.raddr)
    wire(circ.CLK, coreir_mem.clk)
    wire(circ.WADDR, coreir_mem.waddr)
    wire(circ.WDATA, coreir_mem.wdata)
    wire(circ.WE, coreir_mem.wen)
    
    #Register chain for memory
    rd_out = coreir_mem.rdata
    for i in range(read_latency):
        rd_out = register(rd_out)
    wire(circ.RDATA, rd_out)
    
    EndDefine()
    return circ

@lru_cache(maxsize=None)
def DefineMemory(height, width, readonly=False, read_latency=0):
    if read_latency < 0:
        raise ValueError("read_latency cannot be negative")
    if readonly:
        return DefineROM(height, width, read_latency)
    else:
        return DefineRAM(height, width, read_latency)



from mantle import FF
from mantle import LUT2, I0, I1

__all__ = ['rising', 'falling']

def rising(a):
    ff = FF()
    return LUT2(I0 & ~I1)(a, ff(a))


def falling(a):
    ff = FF()
    return LUT2(~I0 & I1)(a, ff(a))



from magma import *
from magma.bitutils import log2
from mantle import Or

__all__ = ['DefineEncoder', 'Encoder', 'encoder']

#
# Given an n-bit input array with only a single bit set,
# return the position of the set bit.
#
# NB. The current implementation only works for n<=8
#
def DefineEncoder(n):
    assert n <= 8
    logn = log2(n)
    class _Encoder(Circuit):
        name = 'Encoder'+str(n)
        IO = ['I', In(Bits[ n ]), 'O', Out(Bits[ logn ])]
        @classmethod
        def definition(Enc):
            def f(y):
                or_ = uncurry(Or(n//2))
                os = []
                for i in range(n):
                    if i & (1 << y): os.append(Enc.I[i])
                wire(array(os), or_)
                return AnonymousCircuit("O", or_.O)
            enc = join( col(f, logn) )
            wire(enc.O, Enc.O)
    return _Encoder

def Encoder(n, **kwargs):
    return DefineEncoder(n)(**kwargs)

def encoder(I, **kwargs):
    return Encoder(len(I), **kwargs)(I)



from collections.abc import Sequence
from magma import *
from ..ice40.PLB import *
from .LUT import LUT2, LUT3
from .logic import Not

__all__   = ['DefineDFF', 'DFF']
__all__  += ['SRFF', 'RSFF', 'JKFF', 'TFF']
__all__  += ['FF']

def DefineDFF(has_ce=False, has_reset=False, edge=True, sync=True):
    # has_set not supported
    has_set = False

    # By default
    #  not connecting a wire to D defaults to 0
    #  not connecting a wire to C defaults to 0
    #  not connecting a wire to R defaults to 0
    #  not connecting a wire to E defaults to 1
    #   this is better than connecting a 1 to E,
    #   which causes that signal to be generated
    if edge:
        # rising edge
        if sync:
            # synchronous reset
            if has_ce:
                if   has_reset:
                    return SB_DFFESR
                elif has_set:
                    return SB_DFFESS
                else:
                    return SB_DFFE
            else:
                if   has_reset:
                    return SB_DFFSR
                elif has_set:
                    return SB_DFFSS
                else:
                    return SB_DFF
        else:
            # asynchronous reset
            if has_ce:
                if   has_reset:
                    return SB_DFFER
                elif has_s:
                    return SB_DFFES
                else:
                    return SB_DFFE
            else:
                if   has_reset:
                    return SB_DFFR
                elif has_set:
                    return SB_DFFS
                else:
                    return SB_DFF
    else:
        # falling edge
        if sync:
            # synchronous reset
            if has_ce:
                if   has_reset:
                    return SB_DFFNESR
                elif has_set:
                    return SB_DFFNESS
                else:
                    return SB_DFFNE
            else:
                if   has_reset:
                    return SB_DFFNSR
                elif has_set:
                    return SB_DFFNSS
                else:
                    return SB_DFFN
        else:
            # asynchronous reset
            if has_ce:
                if   has_reset:
                    return SB_DFFNER
                elif has_set:
                    return SB_DFFNES
                else:
                    return SB_DFFNE
            else:
                if   has_reset:
                    return SB_DFFNR
                elif has_set:
                    return SB_DFFNS
                else:
                    return SB_DFFN

#
# TODO: add async=True, edge=True (also negedge)
#
def DFF(init=0, has_ce=False, has_reset=False, edge=True, sync=True, **kwargs):

    ff = DefineDFF(has_ce, has_reset, edge, sync)(**kwargs)

    I = ff.D
    O = ff.Q

    # ice40 flip-flops are always initialized to 0
    if init:
        # if init=1, then insert Not before and after the flip-flop
        luti = Not()
        luto = Not()
        wire(luti.O, ff.D)
        wire(ff.Q, luto.I0)
        I = luti.I0
        O = luto.O

    args = ['I', I, 'CLK', ff.C]
    if has_ce:
        args += ['CE', ff.E]
    if has_reset:
        args += ['RESET', ff.R]
    #if has_set:
    #    args += ['SET', ff.S]

    args += ['O', O]

    return AnonymousCircuit(args)

FF = DFF


def SRFF(init=0, has_ce=False, edge=True, sync=True, **kwargs):

    """A S-R flip-flop."""

    dff = FF( init=init, has_ce=has_ce, edge=edge, sync=sync, **kwargs)
    lut = LUT3( (~I1&~I2&I0)|(I1&~I2), **kwargs )
    dff(lut)

    wire(dff.O, lut.I0)

    args = []
    if has_ce:
        args += ['CE', dff.CE]
    args += ["S", lut.I1, "R", lut.I2, 'CLK', dff.CLK, "O", dff.O]

    return AnonymousCircuit(args)



def RSFF(init=0, has_ce=False, edge=True, sync=True, **kwargs):

    """A R-S flip-flop."""

    dff = FF( init=init, has_ce=has_ce, edge=edge, sync=sync, **kwargs)
    lut = LUT3( (~I1&~I2&I0)|(I1&~I2), **kwargs )
    dff(lut)

    wire(dff.O, lut.I0)

    args = []
    if has_ce:
        args += ['CE', dff.CE]
    args += ["R", lut.I2, "S", lut.I1, 'CLK', dff.CLK, "O", dff.O]

    return AnonymousCircuit(args)


def JKFF(has_ce=False, has_reset=False, edge=True, sync=True, **kwargs):

    """A J-K flip-flop."""

    dff = FF(has_ce=has_ce, has_reset=has_reset, edge=edge, sync=sync, **kwargs)
    lut = LUT3( (~I0&I1)|(I0&~I2), **kwargs )
    dff(lut)

    wire(dff.O, lut.I0)

    args = ["J", lut.I1, "K", lut.I2, "O", dff.O, 'CLK', dff.CLK]
    if has_ce:     args += ['CE', dff.CE]
    if has_reset:  args += ['RESET', dff.R]
    #if has_set:    args += ['SET', dff.S]
    return AnonymousCircuit(*args)


def TFF(has_ce=False, has_reset=False, edge=True, sync=True, **kwargs):

    """A T flip-flop."""

    tff = FF(has_ce=has_ce, has_reset=has_reset, edge=edge, sync=sync, **kwargs)
    lut = LUT2( I0^I1, **kwargs )
    tff(lut)

    wire(tff.O, lut.I0)

    args = ["I", lut.I1, "O", tff.O, "CLK", tff.CLK]
    if has_ce:    args += ['CE', dff.CE]
    if has_reset: args += ['RESET', dff.R]
    #if has_set:   args += ['SET', dff.S]
    return AnonymousCircuit(*args)




from magma import compose
from .swap import swap, Swaps, swaps
from .permute import Riffle, UnRiffle, ReverseRiffle, UnReverseRiffle

__all__   = ['HalfCleaner', 'halfcleaner']
__all__  += ['ReverseHalfCleaner', 'reversehalfcleaner']

# HalfCleaner
#
# Convert bitonic input into two bitonic halves
#
# Either the top half or the bottom half will be clean, all 0's or 1's
#
def HalfCleaner(n):
    if n == 2:
        return Swap()
    u = UnRiffle(n)
    s = Swaps(n)
    r = Riffle(n)
    return compose( u, compose( s, r ) )

def halfcleaner(I):
    return HalfCleaner(len(I))(I)

#
# ReverseHalfCleaner
#
# Convert 2 sorted inputs into two bitonic halves
#
def ReverseHalfCleaner(n):
    if n == 2:
        return Swap()
    u = UnReverseRiffle(n)
    s = Swaps(n)
    r = ReverseRiffle(n)
    return compose( u, compose( s, r ) )

def reversehalfcleaner(I):
    return ReverseHalfCleaner(len(I))(I)



from magma import Circuit, Bits, In, Out, bits, wire, cache_definition
from magma.bitutils import log2
from .compressor import compressor

__all__ = ['DefinePopCount', 'PopCount', 'popcount']

def DefinePopCount(n):
    class _PopCount(Circuit):
        name = 'PopCount{}'.format(n)
        IO = ['I', In(Bits[ n ]), 'O', Out(Bits[ log2(n)+1 ])]
        @classmethod
        def definition(io):
            r = compressor([io.I.as_list()])
            wire( bits(r), io.O )
    return _PopCount

def PopCount(n, **kwargs):
    return DefinePopCount(n)(**kwargs)

def popcount(I, **kwargs):
    return PopCount(len(I), **kwargs)(I)



from __future__ import division
import sys
if sys.version_info > (3, 0):
    from functools import reduce
    from functools import lru_cache
from collections.abc import Sequence
from magma import *
from .LUT import LUT, LUT1, LUT2, LUT3, LUT4, A0, A1, A2, A3

# unary operators
__all__  = ['DefineReduceAnd', 'ReduceAnd']
__all__ += ['DefineReduceNAnd', 'ReduceNAnd']
__all__ += ['DefineReduceOr', 'ReduceOr']
__all__ += ['DefineReduceNOr', 'ReduceNOr']
__all__ += ['DefineReduceXOr', 'ReduceXOr']
__all__ += ['DefineReduceNXOr', 'ReduceNXOr']

# binary operators
__all__ += ['DefineAnd', 'And']
__all__ += ['DefineNAnd', 'NAnd']
__all__ += ['DefineOr', 'Or']
__all__ += ['DefineNOr', 'NOr']
__all__ += ['DefineXOr', 'XOr']
__all__ += ['DefineNXOr', 'NXOr']

# unary operators
__all__ += ['DefineInvert', 'Invert']
__all__ += ['Not']

def FlatCascade(n, k, expr, cin, **kwargs):

        def f(y):
            e = expr[y] if isinstance(expr, Sequence) else expr
            return LUT( e, n=k+1 )

        # number of luts
        m = (n+k-1) // k
        c = braid( col(f, m), foldargs={"I0":"O"})

        wire(cin, c.I0)

        c = flat(uncurry(c))

        for i in range(n, len(c.I)):
            wire(cin, c.I[i])

        return AnonymousCircuit( ['I', c.I[0:n], 'O', c.O] )

def DefineReduceOp(opname, n, luts, cascadeexpr, cin):
    T = Bits[ n ]
    class _ReduceOp(Circuit):
        name = '{}{}'.format(opname, n)
        IO = ['I', In(T), 'O', Out(Bit)]

        @classmethod
        def definition(io):
            if   n == 1: a = uncurry(LUT1(luts[n - 1]))
            elif n == 2: a = uncurry(LUT2(luts[n - 1]))
            elif n == 3: a = uncurry(LUT3(luts[n - 1]))
            elif n == 4: a = uncurry(LUT4(luts[n - 1]))
            else: a = FlatCascade(n, 1, cascadeexpr, cin)
            wire(a(io.I), io.O)
    return _ReduceOp

def DefineReduceAnd(n):
    luts = [A0, A0&A1, A0&A1&A2, A0&A1&A2&A3]
    return DefineReduceOp('And', n, luts, A0 & A1, 1)

def ReduceAnd(height=2, **kwargs):
    return DefineReduceAnd(height)(**kwargs)

def DefineReduceNAnd(n):
    luts = [~A0, ~(A0&A1), ~(A0&A1&A2), ~(A0&A1&A2&A3)]
    return DefineReduceOp('NAnd', n, luts, A0 & ~A1, 0)

def ReduceNAnd(height=2, **kwargs):
    return DefineReduceNAnd(height)(**kwargs)

def DefineReduceOr(n):
    luts = [A0, A0|A1, A0|A1|A2, A0|A1|A2|A3]
    return DefineReduceOp('Or', n, luts, A0 | A1, 0)

def ReduceOr(height=2, **kwargs):
    return DefineReduceOr(height)(**kwargs)

def DefineReduceNOr(n):
    luts = [~A0, ~(A0|A1), ~(A0|A1|A2), ~(A0|A1|A2|A3)]
    return DefineReduceOp('NOr', n, luts, A0 | ~A1, 1)

def ReduceNOr(height=2, **kwargs):
    return DefineReduceNOr(height)(**kwargs)

def DefineReduceXOr(n):
    luts = [A0, A0^A1, A0^A1^A2, A0^A1^A2^A3]
    return DefineReduceOp('XOr', n, luts, A0 ^ A1, 0)

def ReduceXOr(height=2, **kwargs):
    return DefineReduceXOr(height)(**kwargs)

def DefineReduceNXOr(n):
    luts = [~A0, ~(A0^A1), ~(A0^A1^A2), ~(A0^A1^A2^A3)]
    return DefineReduceOp('NXOr', n, luts, A0 ^ ~A1, 1)

def ReduceNXOr(height=2, **kwargs):
    return DefineReduceNXOr(height)(**kwargs)


def DefineOp(opname, op, height=2, width=1):
    """
    Generate Op module

    I0 : In(Bits(width)), I1 : In(Bits(width)), O : Out(Bits(width))
    """
    T = Bits[ width ]
    class _Op(Circuit):

        name = '{}{}x{}'.format(opname, height, width)

        IO = sum([['I{}'.format(i), In(T)] for i in range(height)], [])
        IO  += ['O', Out(T)]

        @classmethod
        def definition(io):
            def opm(y):
                return curry(op(height))
            opmxn = join(col(opm, width))
            for port in ('I{}'.format(i) for i in range(height)):
                wire(getattr(io, port), getattr(opmxn, port))
            wire(opmxn.O, io.O)
    return _Op

def DefineAnd(height=2, width=1):
    return DefineOp('And', ReduceAnd, height, width)

def And(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceAnd(height, **kwargs))
    return DefineAnd(height, width)(**kwargs)

def DefineNAnd(height=2, width=None):
    return DefineOp('NAnd', ReduceNAnd, height, width)

def NAnd(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceNAnd(height, **kwargs))
    return DefineNAnd(height, width)(**kwargs)

def DefineOr(height=2, width=None):
    return DefineOp('Or', ReduceOr, height, width)

def Or(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceOr(height, **kwargs))
    return DefineOr(height, width)(**kwargs)

def DefineNOr(height=2, width=None):
    return DefineOp('NOr', ReduceNOr, height, width)

def NOr(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceNOr(height, **kwargs))
    return DefineNOr(height, width)(**kwargs)

def DefineXOr(height=2, width=None):
    return DefineOp('XOr', ReduceXOr, height, width)

def XOr(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceXOr(height, **kwargs))
    return DefineXOr(height, width)(**kwargs)

def DefineNXOr(height=2, width=None):
    return DefineOp('NXOr', ReduceNXOr, height, width)

def NXOr(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceNXOr(height, **kwargs))
    return DefineNXOr(height, width)(**kwargs)


def DefineInvert(width):
    """
    Generate Invert module

    I0 : Bits(width) -> O : Bits(width)
    """

    T = Bits[ width ]
    class _Invert(Circuit):
        name = 'Invert%d' % width

        IO  = ['I', In(T), 'O', Out(T)]

        @classmethod
        def definition(def_):
            def not_(y):
                return Not(loc=(0,y/8, y%8))
            invert = join(col(not_, width))
            wire(def_.I, invert.I0)
            wire(invert.O, def_.O)

    return _Invert

def Invert(n, **kwargs):
    return DefineInvert(n)(**kwargs)


def Not(**kwargs):
    """Not gate - 1-bit input."""
    return LUT1(~A0, **kwargs)





from magma.frontend.coreir_ import DefineCircuitFromGeneratorWrapper, GetCoreIRBackend
from magma import cache_definition

@cache_definition
def DefineTerm(width: int):
    """
    Take in an array of wires and connect it to nothing so that you don't get an unconnected
    Returns a circuit definition.

    :param cirb: The CoreIR backend currently be used
    :param width: The width of the element to absorb
    :return:
    """
    return DefineCircuitFromGeneratorWrapper(GetCoreIRBackend(), "coreir", "term", "term_w" + str(width), ["global"],
                                               {"width": width})
def Term(width: int):
    """
    Take in an array of wires and connect it to nothing so that you don't get an unconnected
    Returns an instance of a circuit definition.

    :param cirb: The CoreIR backend currently be used
    :param width: The width of the element to absorb
    :return:
    """
    return DefineTerm(width)()



from magma import DefineCircuit, EndCircuit, In, Out, Bit, Clock

__all__ = ['DFF', 'FF']

_DFF = DefineCircuit('DFF', "D", In(Bit), "CLK", In(Clock), "Q", Out(Bit) )
_DFF.verilog = '''\
    always @(posedge CLK) begin
        Q <= D;
    end'''
EndCircuit()

def DFF(init=0, has_ce=False, has_reset=False, has_set=False):
    ff = _DFF()

    #args = ['I', ff.D, 'CLK', ff.CLK]
    #if has_ce:
    #    args += ['CE', ff.E]
    #if has_reset:
    #    args += ['RESET', ff.R]
    #if has_set:
    #    args += ['SET', ff.S]
    #args += ['O', ff.Q]
    #return AnonymousCircuit(args)

    return ff

FF = DFF



from magma import *
from .register import _RegisterName, FFs

__all__  = ['DefineRing', 'Ring']

def DefineRing(n, init=1, has_ce=False, has_reset=False):
    """
    Generate a n-bit ring counter.

    O : Out(Bits(n))
    """
    class Ring(Circuit):
        name = _RegisterName('Ring', n, init, has_ce, has_reset)
        IO = ['O', Out(Bits[ n ])] + ClockInterface(has_ce,has_reset)
        @classmethod
        def definition(ring):
            ffs = FFs(n, init=init, has_ce=has_ce, has_reset=has_reset)
            reg = scan(ffs, scanargs={"I":"O"})
            reg(reg.O[n-1])
            wire(reg.O, ring.O)
            wireclock(ring, reg)
    return Ring

def Ring(n, init=1, has_ce=False, has_reset=False, **kwargs):
    return DefineRing(n, init, has_ce, has_reset)(**kwargs)



from magma import *
from .register import _RegisterName, FFs

__all__ = ['DefineSISO', 'SISO']

def DefineSISO(n, init=0, has_ce=False, has_reset=False):
    """
    Generate Serial-In, Serial-Out shift register.

    I : In(Bit), O : Out(Bit)
    """

    class _SISO(Circuit):
        name = _RegisterName('SISO', n, init, has_ce, has_reset)
        IO = ['I', In(Bit), 'O', Out(Bit)] + \
               ClockInterface(has_ce,has_reset)
        @classmethod
        def definition(siso):
            ffs = FFs(n, init=init, has_ce=has_ce, has_reset=has_reset)
            reg = braid(ffs, foldargs={"I":"O"})
            reg(siso.I)
            wire(reg.O, siso.O)
            wireclock(siso, reg)
            wiredefaultclock(siso, reg)
    return _SISO

def SISO(n, init=0, has_ce=False, has_reset=False, **kwargs):
    return DefineSISO(n, has_ce=has_ce, has_reset=has_reset)(**kwargs)



from magma import curry
from mantle import DefineReduceAnd, DefineReduceNAnd
from mantle import DefineReduceOr, DefineReduceNOr
from mantle import DefineReduceXOr, DefineReduceNXOr
from mantle import DefineAnd, DefineNAnd
from mantle import DefineOr, DefineNOr
from mantle import DefineXOr, DefineNXOr
from mantle import DefineInvert

def ReduceAnd(height=2, **kwargs):
    return DefineReduceAnd(height)(**kwargs)

def ReduceNAnd(height=2, **kwargs):
    return DefineReduceNAnd(height)(**kwargs)

def ReduceOr(height=2, **kwargs):
    return DefineReduceOr(height)(**kwargs)

def ReduceNOr(height=2, **kwargs):
    return DefineReduceNOr(height)(**kwargs)

def ReduceXOr(height=2, **kwargs):
    return DefineReduceXOr(height)(**kwargs)

def ReduceNXOr(height=2, **kwargs):
    return DefineReduceNXOr(height)(**kwargs)


def And(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceAnd(height, **kwargs))
    return DefineAnd(height, width)(**kwargs)

def NAnd(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceNAnd(height, **kwargs))
    return DefineNAnd(height, width)(**kwargs)

def Or(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceOr(height, **kwargs))
    return DefineOr(height, width)(**kwargs)

def NOr(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceNOr(height, **kwargs))
    return DefineNOr(height, width)(**kwargs)

def XOr(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceXOr(height, **kwargs))
    return DefineXOr(height, width)(**kwargs)

def NXOr(height=2, width=None, **kwargs):
    if width is None:
        return curry(ReduceNXOr(height, **kwargs))
    return DefineNXOr(height, width)(**kwargs)

def Invert(n, **kwargs):
    return DefineInvert(n)(**kwargs)




#
# Implementation of compressor trees
#
from magma import fork, cache_definition, DefineCircuit, EndDefine, In, Out, Bit, wire
from magma import mantle_target
if mantle_target == "ice40":
    from mantle import LUT2, LUT3, I0, I1, I2
else:
    from mantle import XOr, Or, And


__all__ = ['compressor']

n3to2s = 0
n2to2s = 0

if mantle_target == "ice40":
    def compress2to2():
        global n2to2s
        n2to2s = n2to2s + 1
        return fork([LUT2(I0^I1), LUT2((I0&I1)|(I1&I2))])

    def compress3to2():
        global n3to2s
        n3to2s = n3to2s + 1
        return fork([LUT3(I0^I1^I2), LUT3((I0&I1)|(I1&I2)|(I2&I0))])
else:
    def Define3to2Op():
        Op = DefineCircuit("_3to2Op", "I0", In(Bit), "I1", In(Bit), "I2", In(Bit), "O", Out(Bit))
        a = And(2)(Op.I0, Op.I1)
        b = And(2)(Op.I1, Op.I2)
        c = And(2)(Op.I2, Op.I0)
        d = Or(3)(a, b, c)
        wire(d, Op.O)
        EndDefine()
        return Op

    def compress3to2():
        global n3to2s
        n3to2s = n3to2s + 1
        Op = Define3to2Op()
        return fork([XOr(3), Op()])

# compress 2 and 3 bit groups in a column
def compresscolumn2(bits):
    n = len(bits)
    ones = []
    twos = []
    for i in range(0,n,3):
        if i+1 < n:
            c = compress3to2()
            c(bits[i], bits[i+1], 0 if i+2 >= n else bits[i+2])
            ones.append(c.O[0])
            twos.append(c.O[1])
        else:
            while i < n:
                ones.append(bits[i])
                i += 1
            break
    return twos, ones

# compress only 3 bit groups in a column
def compresscolumn3(bits):
    n = len(bits)
    ones = []
    twos = []
    for i in range(0,n,3):
        if i+2 < n:
            c = compress3to2()
            c(bits[i], bits[i+1], bits[i+2])
            ones.append(c.O[0])
            twos.append(c.O[1])
        else:
            while i < n:
                ones.append(bits[i])
                i += 1
            break
    return twos, ones


# compress only 3 bit groups in a column without ripple
def compress3(bits):
    res = []
    lasttwos = []
    for b in bits:
        twos, ones = compresscolumn3(b)
        res.append(ones + lasttwos)
        lasttwos = twos
    if len(lasttwos) > 0:
        res.append(lasttwos)
    return res

# ripple adder - assumes all the columns have at most 2 bits
def ripple(bits):
    res = []
    twos = []
    for b in bits:
        # ripple: combine last column of twos with this column
        b = b + twos
        # should only require fulladders (compress 3 to 2)
        assert len(b) <= 3
        twos, ones = compresscolumn2(b)
        res.append(ones)
    if len(twos) > 0:
        res.append(twos)
    return res


# returns True if there are any columns with more than 2 bits
def iscompressible(bits):
    for b in bits:
         if len(b) > 2:
             return True
    return False

def compressor(r):
    global n2to2s, n3to2s
    n3to2s = 0
    n2to2s = 0
    #print( list(map(len, r)) )
    while iscompressible(r):
        r = compress3(r)
        #print( list(map(len, r)) )
    r = ripple(r)
    #print( list(map(len, r)) )
    #print('n3to2s = {}'.format(n3to2s))
    return sum(r, [])




from magma import In, Out, Bits, Circuit, wire, cache_definition

__all__  = ['DefinePermute', 'Permute']
__all__ += ['Reverse']
__all__ += ['Riffle', 'UnRiffle']
__all__ += ['ReverseRiffle', 'UnReverseRiffle']
__all__ += ['EvenOdd']

def flatten(l):
    return sum(l, [])

def shuffle(l, r):
    return flatten([[l[i], r[i]] for i in range(n//2)])

def concat(l, r):
    return l+r

def inverse(l):
    return [l.index(i) for i in range(len(l))]


def _identity(n):
    return range(n)

def _reverse(n):
    return list(reversed(range(n)))

def _even(n):
    return [2*i for i in range(n//2)]

def _odd(n):
    return [2*i+1 for i in range(n//2)]

def _bottom(n):
    return [i for i in range(n//2)]

def _top(n):
    return [n//2+i for i in range(n//2)]

def _evenodd(n):
    return _even(n) + _odd(n)

def _riffle(n):
    return flatten([[i, i+n//2] for i in range(n//2)])

def _unriffle(n):
    return inverse(_riffle(n))

def _reverseriffle(n):
    return flatten([[i, n-1-i] for i in range(n//2)])

def _unreverseriffle(n):
    return inverse(_reverseriffle(n))

def DefinePermute(na,permutation):
    n = len(permutation)
    class Permute(Circuit):
        name = na
        IO = ["I", In(Bits(n)), "O", Out(Bits(n))]
        @classmethod
        def definition(io):
            [wire(io.I[permutation[i]], io.O[i]) for i in range(len(io.I))]
    return Permute

def Permute(name,permutation):
    return DefinePermute(name,tuple(permutation))()


def Reverse(n):
    return Permute('Reverse{}'.format(n),_reverse(n))

def Riffle(n):
    return Permute('Riffle{}'.format(n),_riffle(n))

def UnRiffle(n):
    return Permute('UnRiffle{}'.format(n),_unriffle(n))

def ReverseRiffle(n):
    return Permute('ReverseRiffle{}'.format(n),_reverseriffle(n))

def UnReverseRiffle(n):
    return Permute('UnReverseRiffle{}'.format(n),_unreverseriffle(n))

def EvenOdd(n):
    return Permute('EvenOdd{}'.format(n),_evenodd(n))




import mantle
import magma as m
import importlib


def weak_import(module, scope):
    # Iterate over names defined in module, only define them here if
    # they haven't already been defined. This allows implementations to
    # specialize a common module. For example, mantle40 defines a FullAdder, so
    # it will not use the generic FullAdder in common.
    module = importlib.import_module(f"mantle.{module}")
    if hasattr(module, "__all__"):
        names = module.__all__
    else:
        names = dir(module)
    for name in names:
        try:
            getattr(mantle, name)
        except AttributeError:
            scope.update({name: getattr(module, name)})

if m.mantle_target is None:
    m.set_mantle_target('coreir')

if m.mantle_target in ['coreir',
                       'ice40',
                       'spartan3',
                       'spartan6',
                       'kintex7',
                       'cyclone4',
                       'verilog'
    ]:

    if m.mantle_target == 'coreir':
        from mantle.coreir import *
    elif m.mantle_target in ['ice40']:
        from mantle.lattice import *
    elif m.mantle_target in ['spartan3', 'spartan6', 'kintex7']:
        from mantle.xilinx import *
    elif m.mantle_target in ['cyclone4']:
        from mantle.altera import *
    #elif m.mantle_target in ['greenpak4']:
    #    from mantle.silego import *
    elif m.mantle_target == 'verilog':
        from mantle.verilog import *

    if m.mantle_target == 'verilog':
        print("Warning: verilog mantle target does not support common, will not be imported")
    else:
        import mantle.common
        for module in mantle.common.modules:
            weak_import(f"common.{module}", globals())

else:
    if m.mantle_target not in ['greenpak4']:
        raise RuntimeError(f"MANTLE_TARGET={m.mantle_target} not supported")
